/*
 * Copyright 2021 Fluence Labs Limited
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

use air::ExecutionCidState;
use air_test_utils::prelude::*;

use std::cell::RefCell;

#[test]
fn ap_with_scalars() {
    let vm_1_peer_id = "vm_1_peer_id";
    let test_value = "scalar_2";
    let mut vm_1 = create_avm(set_variable_call_service(json!({ "field": test_value })), vm_1_peer_id);

    let vm_2_peer_id = "vm_2_peer_id";
    let mut vm_2 = create_avm(echo_call_service(), vm_2_peer_id);

    let script = f!(r#"
        (seq
            (seq
                (call "{vm_1_peer_id}" ("" "") ["scalar_1_result"] scalar_1)
                (ap scalar_1.$.field! scalar_2)
            )
            (call "{vm_2_peer_id}" ("" "") [scalar_2])
        )
        "#);

    let result = checked_call_vm!(vm_1, <_>::default(), &script, "", "");
    let result = checked_call_vm!(vm_2, <_>::default(), script, "", result.data);

    let actual_trace = trace_from_result(&result);
    let expected_state = vec![
        scalar!(
            json!({ "field": test_value }),
            peer = vm_1_peer_id,
            args = ["scalar_1_result"]
        ),
        unused!(test_value, peer = vm_2_peer_id, args = [test_value]),
    ];

    assert_eq!(actual_trace, expected_state);
    assert!(result.next_peer_pks.is_empty());
}

#[test]
fn ap_with_string_literal() {
    let vm_1_peer_id = "vm_1_peer_id";
    let mut vm_1 = create_avm(echo_call_service(), vm_1_peer_id);

    let some_string = "some_string";
    let script = f!(r#"
        (seq
            (ap "{some_string}" $stream)
            (seq
                (canon "{vm_1_peer_id}" $stream #canon_stream)
                (call "{vm_1_peer_id}" ("" "") [#canon_stream])))
        "#);

    let result = checked_call_vm!(vm_1, <_>::default(), script, "", "");

    let actual_trace = trace_from_result(&result);
    let expected_state = vec![
        executed_state::ap(0),
        executed_state::canon(json!(
            {
            "tetraplet": {"function_name": "", "json_path": "", "peer_pk": "vm_1_peer_id", "service_id": ""},
            "values": [
                {
                    "result": "some_string",
                    "tetraplet": {"function_name": "", "json_path": "", "peer_pk": "", "service_id": ""},
                    "trace_pos": 0
                }
            ]
        }
        )),
        unused!(json!([some_string]), peer = vm_1_peer_id, args = [json!([some_string])]),
    ];

    assert_eq!(actual_trace, expected_state);
    assert!(result.next_peer_pks.is_empty());
}

#[test]
fn ap_with_bool_literal() {
    let vm_1_peer_id = "vm_1_peer_id";
    let mut vm_1 = create_avm(echo_call_service(), vm_1_peer_id);

    let script = f!(r#"
        (seq
            (ap true $stream)
            (seq
                (canon "{vm_1_peer_id}" $stream #canon_stream)
                (call "{vm_1_peer_id}" ("" "") [#canon_stream])))
        "#);

    let result = checked_call_vm!(vm_1, <_>::default(), script, "", "");

    let actual_trace = trace_from_result(&result);
    let expected_state = vec![
        executed_state::ap(0),
        executed_state::canon(json!(   {
            "tetraplet": {"function_name": "", "json_path": "", "peer_pk": "vm_1_peer_id", "service_id": ""},
            "values": [
                {
                    "result": true,
                    "tetraplet": {"function_name": "", "json_path": "", "peer_pk": "", "service_id": ""},
                    "trace_pos": 0
                }
            ]
        })),
        unused!(json!([true]), peer = vm_1_peer_id, args = [json!([true])]),
    ];

    assert_eq!(actual_trace, expected_state);
    assert!(result.next_peer_pks.is_empty());
}

#[test]
fn ap_with_number_literal() {
    let vm_1_peer_id = "vm_1_peer_id";
    let mut vm_1 = create_avm(echo_call_service(), vm_1_peer_id);

    let script = f!(r#"
        (seq
            (ap 100 $stream)
            (seq
                (canon "{vm_1_peer_id}" $stream #canon_stream)
                (call "{vm_1_peer_id}" ("" "") [#canon_stream])))
        "#);

    let result = checked_call_vm!(vm_1, <_>::default(), script, "", "");

    let actual_trace = trace_from_result(&result);
    let expected_state = vec![
        executed_state::ap(0),
        executed_state::canon(json!({
            "tetraplet": {"function_name": "", "json_path": "", "peer_pk": "vm_1_peer_id", "service_id": ""},
            "values": [
                {
                    "result": 100,
                    "tetraplet": {"function_name": "", "json_path": "", "peer_pk": "", "service_id": ""},
                    "trace_pos": 0
                }
            ]
        })),
        unused!(json!([100]), peer = vm_1_peer_id, args = [json!([100])]),
    ];

    assert_eq!(actual_trace, expected_state);
    assert!(result.next_peer_pks.is_empty());
}

#[test]
fn ap_with_last_error() {
    let vm_1_peer_id = "vm_1_peer_id";
    let mut vm_1 = create_avm(echo_call_service(), vm_1_peer_id);

    let script = f!(r#"
        (seq
            (ap %last_error% $stream)
            (seq
                (canon "{vm_1_peer_id}" $stream #canon_stream)
                (call "{vm_1_peer_id}" ("" "") [#canon_stream])))
        "#);

    let result = checked_call_vm!(vm_1, <_>::default(), script, "", "");

    let actual_trace = trace_from_result(&result);
    let expected_state = vec![
        executed_state::ap(0),
        executed_state::canon(json!({
            "tetraplet": {"function_name": "", "json_path": "", "peer_pk": "vm_1_peer_id", "service_id": ""},
            "values": [
                {
                    "result": null,
                    "tetraplet": {"function_name": "", "json_path": "", "peer_pk": "", "service_id": ""},
                    "trace_pos": 0
                }
            ]
        })),
        unused!(json!([null]), peer = vm_1_peer_id, args = [json!([null])]),
    ];

    assert_eq!(actual_trace, expected_state);
    assert!(result.next_peer_pks.is_empty());
}

#[test]
fn ap_with_timestamp() {
    let vm_1_peer_id = "vm_1_peer_id";
    let mut vm_1 = create_avm(echo_call_service(), vm_1_peer_id);

    let script = f!(r#"
        (seq
            (ap %timestamp% scalar)
            (call "{vm_1_peer_id}" ("" "") [scalar])
        )
        "#);

    let test_params = TestRunParameters::from_timestamp(1337);
    let result = checked_call_vm!(vm_1, test_params.clone(), script, "", "");

    let actual_trace = trace_from_result(&result);
    let expected_state = vec![unused!(
        test_params.timestamp,
        peer = vm_1_peer_id,
        args = [test_params.timestamp]
    )];

    assert_eq!(actual_trace, expected_state);
}

#[test]
fn ap_with_ttl() {
    let vm_1_peer_id = "vm_1_peer_id";
    let mut vm_1 = create_avm(echo_call_service(), vm_1_peer_id);

    let script = f!(r#"
        (seq
            (ap %ttl% scalar)
            (call "{vm_1_peer_id}" ("" "") [scalar])
        )
        "#);

    let test_params = TestRunParameters::from_ttl(1337);
    let result = checked_call_vm!(vm_1, test_params.clone(), script, "", "");

    let actual_trace = trace_from_result(&result);
    let expected_state = vec![unused!(test_params.ttl, peer = vm_1_peer_id, args = [test_params.ttl])];

    assert_eq!(actual_trace, expected_state);
}

#[test]
fn ap_with_dst_stream() {
    let vm_1_peer_id = "vm_1_peer_id";
    let test_value = "scalar_2";
    let mut vm_1 = create_avm(set_variable_call_service(json!({ "field": test_value })), vm_1_peer_id);

    let vm_2_peer_id = "vm_2_peer_id";
    let mut vm_2 = create_avm(echo_call_service(), vm_2_peer_id);

    let script = f!(r#"
        (seq
            (seq
                (call "{vm_1_peer_id}" ("" "") ["scalar_1_result"] scalar_1)
                (ap scalar_1 $stream))
            (seq
                (canon "{vm_2_peer_id}" $stream #canon_stream)
                (call "{vm_2_peer_id}" ("" "") [#canon_stream])))
        "#);

    let result = checked_call_vm!(vm_1, <_>::default(), &script, "", "");
    let result = checked_call_vm!(vm_2, <_>::default(), script, "", result.data);

    let actual_trace = trace_from_result(&result);
    let expected_state = vec![
        scalar!(
            json!({ "field": test_value }),
            peer = vm_1_peer_id,
            args = ["scalar_1_result"]
        ),
        executed_state::ap(0),
        executed_state::canon(json!({
            "tetraplet": {"function_name": "", "json_path": "", "peer_pk": "vm_2_peer_id", "service_id": ""},
            "values": [
                {
                    "result": {"field": "scalar_2"},
                    "tetraplet": {"function_name": "", "json_path": "", "peer_pk": "vm_1_peer_id", "service_id": ""},
                    "trace_pos": 1
                }
            ]
        })),
        unused!(
            json!([{ "field": test_value }]),
            peer = vm_2_peer_id,
            args = [json!([{ "field": test_value }])]
        ),
    ];

    assert_eq!(actual_trace, expected_state);
    assert!(result.next_peer_pks.is_empty());
}

#[test]
fn ap_canon_stream_with_lambda() {
    let vm_1_peer_id = "vm_1_peer_id";
    let (echo_call_service, tetraplet_checker) = tetraplet_host_function(echo_call_service());
    let mut vm_1 = create_avm(echo_call_service, vm_1_peer_id);

    let service_name = "some_service_name";
    let function_name = "some_function_name";
    let script = f!(r#"
        (seq
            (seq
                (call "{vm_1_peer_id}" ("" "") [0] $stream)
                (call "{vm_1_peer_id}" ("{service_name}" "{function_name}") [1] $stream))
            (seq
                (canon "{vm_1_peer_id}" $stream #canon_stream)
                (seq
                    (ap #canon_stream.$.[1] $stream_2)
                    (seq
                        (canon "{vm_1_peer_id}" $stream_2 #canon_stream_2)
                        (call "{vm_1_peer_id}" ("" "") [#canon_stream_2])))))
        "#);

    let result = checked_call_vm!(vm_1, <_>::default(), &script, "", "");

    let actual_trace = trace_from_result(&result);
    let expected_state = vec![
        stream!(0, 0, peer = vm_1_peer_id, args = [0]),
        stream!(
            1,
            1,
            peer = vm_1_peer_id,
            service = service_name,
            function = function_name,
            args = [1]
        ),
        executed_state::canon(
            json!({"tetraplet": {"function_name": "", "json_path": "", "peer_pk": "vm_1_peer_id", "service_id": ""},
                "values": [{"result": 0, "tetraplet": {"function_name": "", "json_path": "", "peer_pk": "vm_1_peer_id", "service_id": ""}, "trace_pos": 0},
                    {"result": 1, "tetraplet": {"function_name": "some_function_name", "json_path": "", "peer_pk": "vm_1_peer_id", "service_id": "some_service_name"}, "trace_pos": 1}]}),
        ),
        executed_state::ap(0),
        executed_state::canon(json!(
            {
                "tetraplet": {"function_name": "", "json_path": "", "peer_pk": "vm_1_peer_id", "service_id": ""},
                "values": [
                    {
                        "result": 1,
                        "tetraplet": {"function_name": "some_function_name", "json_path": ".$.[1]", "peer_pk": "vm_1_peer_id", "service_id": "some_service_name"},
                        "trace_pos": 3,
                    }
            ]
            }
        )),
        unused!(json!([1]), peer = vm_1_peer_id, args = [json!([1])]),
    ];
    assert_eq!(actual_trace, expected_state);

    let expected_tetraplet = RefCell::new(vec![vec![SecurityTetraplet::new(
        vm_1_peer_id,
        service_name,
        function_name,
        ".$.[1]",
    )]]);
    assert_eq!(tetraplet_checker.as_ref(), &expected_tetraplet);
}

#[test]
fn ap_canon_stream() {
    let vm_1_peer_id = "vm_1_peer_id";
    let (echo_call_service, tetraplet_checker) = tetraplet_host_function(echo_call_service());
    let mut vm_1 = create_avm(echo_call_service, vm_1_peer_id);

    let service_name = "some_service_name";
    let function_name = "some_function_name";
    let script = f!(r#"
        (seq
            (seq
                (call "{vm_1_peer_id}" ("" "") [0] $stream)
                (call "{vm_1_peer_id}" ("{service_name}" "{function_name}") [1] $stream))
            (seq
                (canon "{vm_1_peer_id}" $stream #canon_stream)
                (seq
                    (ap #canon_stream $stream_2)
                    (seq
                        (canon "{vm_1_peer_id}" $stream_2 #canon_stream_2)
                        (call "{vm_1_peer_id}" ("" "") [#canon_stream_2])))))
        "#);

    let result = checked_call_vm!(vm_1, <_>::default(), &script, "", "");
    print_trace(&result, "");

    let actual_trace = trace_from_result(&result);
    let expected_state = vec![
        stream!(0, 0, peer = vm_1_peer_id, args = [0]),
        stream!(
            1,
            1,
            peer = vm_1_peer_id,
            service = service_name,
            function = function_name,
            args = [1]
        ),
        executed_state::canon(
            json!({"tetraplet": {"function_name": "", "json_path": "", "peer_pk": "vm_1_peer_id", "service_id": ""},
                "values": [{"result": 0, "tetraplet": {"function_name": "", "json_path": "", "peer_pk": "vm_1_peer_id", "service_id": ""}, "trace_pos": 0},
                    {"result": 1, "tetraplet": {"function_name": "some_function_name", "json_path": "", "peer_pk": "vm_1_peer_id", "service_id": "some_service_name"}, "trace_pos": 1}]}),
        ),
        executed_state::ap(0),
        executed_state::canon(json!(
            {
                "tetraplet": {"function_name": "", "json_path": "", "peer_pk": "vm_1_peer_id", "service_id": ""},
                "values": [
                    {
                        "result": [0, 1],
                        "tetraplet": {"function_name": "", "json_path": "", "peer_pk": "vm_1_peer_id", "service_id": ""},
                        "trace_pos": 3,
                    }
            ]
            }
        )),
        unused!(json!([[0, 1]]), peer = vm_1_peer_id, args = [json!([[0, 1]])]),
    ];
    assert_eq!(actual_trace, expected_state);

    let expected_tetraplet = RefCell::new(vec![vec![SecurityTetraplet::new(vm_1_peer_id, "", "", "")]]);
    assert_eq!(tetraplet_checker.as_ref(), &expected_tetraplet);
}

#[test]
fn ap_stream_map() {
    let vm_1_peer_id = "vm_1_peer_id";
    let mut vm_1 = create_avm(echo_call_service(), vm_1_peer_id);

    let service_name1 = "serv1";
    let service_name2 = "serv2";
    let script = f!(r#"
        (seq
            (seq
                (ap "{vm_1_peer_id}" "{service_name1}" %map)
                (ap "{vm_1_peer_id}" "{service_name2}" %map)
            )
            (fold %map i
                (seq
                    (call i.$.key (i.$.key i.$.value) [i] u)
                    (next i)
                )
            )
        )
        "#);

    let result = checked_call_vm!(vm_1, <_>::default(), &script, "", "");
    let actual_trace = trace_from_result(&result);
    let generation_idx = 0;
    let mut cid_tracker = ExecutionCidState::new();
    let service_result1 = json!({
      "key": vm_1_peer_id,
      "value": service_name1,
    });
    let service_result2 = json!({
      "key": vm_1_peer_id,
      "value": service_name2,
    });
    let service_args1 = vec![service_result1.clone()];
    let service_args2 = vec![service_result2.clone()];

    let expected_state = ExecutionTrace::from(vec![
        executed_state::ap(generation_idx),
        executed_state::ap(generation_idx),
        executed_state::fold(vec![
            subtrace_lore(0, SubTraceDesc::new(3.into(), 1), SubTraceDesc::new(5.into(), 0)),
            subtrace_lore(1, SubTraceDesc::new(4.into(), 1), SubTraceDesc::new(5.into(), 0)),
        ]),
        scalar_tracked!(
            service_result1,
            cid_tracker,
            peer = vm_1_peer_id,
            service = vm_1_peer_id,
            function = service_name1,
            args = service_args1
        ),
        scalar_tracked!(
            service_result2,
            cid_tracker,
            peer = vm_1_peer_id,
            service = vm_1_peer_id,
            function = service_name2,
            args = service_args2
        ),
    ]);
    assert_eq!(actual_trace, expected_state);
}
